# Data Compression Preserving the Region-of-Interest (RoI) using MGARD 

First, build and install MGARD. Then, build the MGARD-RoI example as follows:

```
$ build/cmake ..
$ build/make
```

The `mgard_roi.cpp` shows an example of how to detect the RoI and prescribe region-wise error bound during lossy data compression. Here, we define the RoIs as places where variational features may exist. The code uses MGARD decomposed coefficients and an algorithm similar to Adaptive Mesh Refinement (AMR) for RoI detection. The demonstration takes the simulation data generated by E3SM Atmosphere Model (https://e3sm.org) as the example, compresses it according to the prescribed error bounds placed on RoIs, and reconstructs the data. 
We include 3 variables in the testing dataset, which can be used for tropical cyclone and atmospheric river tracking. The input variables are 3-dimensional -- `{timestep, {x, y}}`. The example code can be parameterized to perform feature tracking and region-adaptive compression in both time-space (3D) and space (2D). When visualize the reconstructed data, it shall have lower errors (bounded by the input error tolerance) in the regions where data are more turbulent and higher errors in the rest regions.

#### Test dataset is available at: https://www.dropbox.com/s/ak0iaypb12d3ojt/dataset.zip?dl=0

Run `run_script.sh` for demonstration or see below for the inputs/outputs of the executable `mgard_roi`: 

* Inputs
   * Path to the input data
   * Filename of the input data 
   * Variable to read in (choose from the testing dataset)
   * Dims (2 or 3 for spatial or spatiotemporal adaptive compression)
   * Number of refinement levels for RoI detection (recommend 1-4) 
   * The coarsest level at which MGARD multilevel coefficients will be used for RoI detection (recommend log2(min{Dim1, Dim2, ..}) - 1)
   * Initial bin width (bw) used for refinement (bw_dim1, bw_dim2, …)
   * Ratios of bin width used before and after each refinement 
   * RoIs as the percentage of data to be kept at each level of refinement
*  Outputs
   * Compression ratios in display
   * Reconstructed data for evaluation 

